@@ -85,21 +85,13 @@ struct _AppMenuBridgePrivate
 
   GDBusProxy       *appmenuproxy;
   gboolean          online;
+
+  GSList           *rebuild_list;
+  guint             rebuild_timer_id;
 };
 
-typedef struct _RecurseContext
-{
-  AppMenuBridge    *bridge;
-  AppWindowContext *context;
-
-  gint count;
-  gboolean previous;
-  DbusmenuMenuitem *stack[30];
-} RecurseContext;
-
 G_DEFINE_DYNAMIC_TYPE(AppMenuBridge, app_menu_bridge, UBUNTU_TYPE_MENU_PROXY)
 
-static GHashTable *rebuild_ids = NULL;
 static GDBusNodeInfo *      registrar_node_info = NULL;
 static GDBusInterfaceInfo * registrar_interface_info = NULL;
 
@@ -213,28 +205,58 @@ context_free (AppWindowContext *context)
   g_free (context);
 }
 
+
+/* a widget that was queued for rebuild has been destroyed,
+   so remove it from our queue */
+static void
+rebuild_list_widget_destroyed (gpointer gbridge, GObject * dead_object)
+{
+  AppMenuBridge * bridge = APP_MENU_BRIDGE (gbridge);
+  AppMenuBridgePrivate * p = bridge->priv;
+
+  p->rebuild_list = g_slist_remove (p->rebuild_list, dead_object);
+}
+
+static void
+rebuild_list_clear (AppMenuBridge * bridge)
+{
+  GSList * l;
+  AppMenuBridgePrivate * p = bridge->priv;
+
+  /* unref the widgets and blow away the list */
+  for (l=p->rebuild_list; l!=NULL; l=l->next)
+    g_object_weak_unref (l->data, rebuild_list_widget_destroyed, bridge);
+  g_slist_free (p->rebuild_list);
+  p->rebuild_list = NULL;
+
+  /* remove the rebuild timer */
+  if (p->rebuild_timer_id > 0)
+    {
+      g_source_remove (p->rebuild_timer_id);
+      p->rebuild_timer_id = 0;
+    }
+}
+
 static void
 app_menu_bridge_dispose (GObject *object)
 {
   AppMenuBridge *bridge = APP_MENU_BRIDGE (object);
+  AppMenuBridgePrivate * p = bridge->priv;
 
-  g_list_foreach (bridge->priv->windows, (GFunc)context_dispose, NULL);
-
-  if (bridge->priv->appmenuproxy)
-    {
-      g_object_unref (bridge->priv->appmenuproxy);
-      bridge->priv->appmenuproxy = NULL;
-    }
+  rebuild_list_clear (bridge);
+  g_list_foreach (p->windows, (GFunc)context_dispose, NULL);
+  g_clear_object (&p->appmenuproxy);
 }
 
 static void
 app_menu_bridge_finalize (GObject *object)
 {
   AppMenuBridge *bridge = APP_MENU_BRIDGE (object);
+  AppMenuBridgePrivate * p = bridge->priv;
 
-  g_list_foreach (bridge->priv->windows, (GFunc)context_free, NULL);
-  g_list_free (bridge->priv->windows);
-  bridge->priv->windows = NULL;
+  g_list_foreach (p->windows, (GFunc)context_free, NULL);
+  g_list_free (p->windows);
+  p->windows = NULL;
 
   G_OBJECT_CLASS (app_menu_bridge_parent_class)->finalize (object);
 }
@@ -505,63 +527,51 @@ appmenuproxy_created_cb (GObject * object, GAsyncResult * res, gpointer user_dat
 	return;
 }
 
-typedef struct _RebuildData {
-  AppMenuBridge *bridge;
-  GtkWidget     *widget;
-} RebuildData;
-
 static gboolean
-do_rebuild (RebuildData *data)
+rebuild_timer_func (gpointer gbridge)
 {
-  if (data->widget != NULL && gtk_widget_is_toplevel (data->widget))
+  GSList * l;
+  AppMenuBridge *bridge = APP_MENU_BRIDGE (gbridge);
+  AppMenuBridgePrivate * p = bridge->priv;
+
+  /* call rebuild_window_items() for each toplevel in our list */
+  for (l=p->rebuild_list; l!=NULL; l=l->next)
     {
-      rebuild_window_items (data->bridge,
-                            data->widget);
+      GtkWidget * w = GTK_WIDGET(l->data);
+
+      if (gtk_widget_is_toplevel(w))
+        rebuild_window_items (bridge, w);
     }
 
-  if (data->widget != NULL)
-    {
-      g_object_remove_weak_pointer (G_OBJECT (data->widget), (gpointer*)&data->widget);
-      g_hash_table_remove (rebuild_ids, data->widget);
-    }
-
-  g_free (data);
-
-  return FALSE;
+  /* cleanup */
+  rebuild_list_clear (bridge);
+  return G_SOURCE_REMOVE;
 }
 
+/* widgets tend to get a lot of changes all at once...
+ + inserting a 100 msec delay here lets us batch up those changes */
 static void
-rebuild (AppMenuBridge *bridge,
-         GtkWidget     *toplevel)
+rebuild (AppMenuBridge * bridge, GtkWidget * toplevel)
 {
-  guint id = 0;
+  AppMenuBridgePrivate * p = bridge->priv;
 
-  if (rebuild_ids != NULL)
+  g_return_if_fail (toplevel != NULL);
+
+  /* ensure there's a rebuild timer ticking */
+  if (p->rebuild_timer_id == 0)
     {
-      id = GPOINTER_TO_UINT (g_hash_table_lookup (rebuild_ids, toplevel));
-
-      if (id > 0)
-        {
-          g_source_remove (id);
-          g_hash_table_remove (rebuild_ids, toplevel);
-          id = 0;
-        }
+      p->rebuild_timer_id = gdk_threads_add_timeout (100,
+                                                     rebuild_timer_func,
+                                                     bridge);
     }
 
-  RebuildData *data = g_new0 (RebuildData, 1);
-  data->bridge = bridge;
-  data->widget = toplevel;
-
-  id = gdk_threads_add_timeout (100, (GSourceFunc)do_rebuild, data);
-
-  g_object_add_weak_pointer (G_OBJECT (data->widget), (gpointer*)&data->widget);
-
-  if (rebuild_ids == NULL)
+  /* ensure this widget is in our rebuild list */
+  if (g_slist_find (p->rebuild_list, toplevel) == NULL)
     {
-      rebuild_ids = g_hash_table_new (g_direct_hash, g_direct_equal);
+      GObject * o = G_OBJECT(toplevel);
+      p->rebuild_list = g_slist_prepend (p->rebuild_list, o);
+      g_object_weak_ref (o, rebuild_list_widget_destroyed, bridge);
     }
-
-  g_hash_table_insert (rebuild_ids, toplevel, GUINT_TO_POINTER (id));
 }
 
 static DbusmenuMenuitem * find_menu_bar (GtkWidget * widget);
@@ -585,6 +595,10 @@ static DbusmenuMenuitem *
 find_menu_bar (GtkWidget * widget)
 {
 	if (GTK_IS_MENU_BAR(widget) || GTK_IS_MENU_ITEM(widget)) {
+                /* If this is from gtkmodelmenu ignore it */
+                if (g_object_get_data (G_OBJECT (widget), "gtk-model-menu-binding"))
+                        return NULL;
+
 		return dbusmenu_gtk_parse_menu_structure(widget);
 	}
 
@@ -822,9 +836,11 @@ attach_notify_cb (GtkWidget     *widget,
 
       g_object_get (widget, "attach-widget", &attach, NULL);
 
-      rebuild (bridge, attach);
-
-      g_clear_object (&attach);
+      if (attach != NULL)
+        {
+          rebuild (bridge, attach);
+          g_object_unref (attach);
+        }
     }
 }
 
@@ -837,6 +853,10 @@ app_menu_bridge_insert (UbuntuMenuProxy *proxy,
   AppMenuBridge        *bridge;
   GtkWidget            *toplevel = NULL;
 
+  /* If this is from gtkmodelmenu, don't bother... */
+  if (g_object_get_data (G_OBJECT (parent), "gtk-model-menu-binding"))
+    return;
+
   if (GTK_IS_TEAROFF_MENU_ITEM (child))
     return;
 
@@ -947,6 +967,7 @@ app_menu_brige_shouldnt_load (void)
       || (g_strcmp0 (prg, "indicator-loader") == 0)
       || (g_strcmp0 (prg, "mutter") == 0)
       || (g_strcmp0 (prg, "midori") == 0)
+      || (g_strcmp0 (prg, "qtoctave") == 0)
       || (g_strcmp0 (prg, "firefox-bin") == 0)
       || (g_strcmp0 (prg, "thunderbird-bin") == 0)
       || (g_strcmp0 (prg, "Eclipse") == 0)
@@ -955,6 +976,7 @@ app_menu_brige_shouldnt_load (void)
       || (g_strcmp0 (prg, "emacs23-lucid") == 0)
       || (g_strcmp0 (prg, "glade") == 0)
       || (g_strcmp0 (prg, "gnome-panel") == 0)
+      || (g_strcmp0 (prg, "wicd-client.py") == 0)
       || (g_strcmp0 (prg, "Lotus Notes") == 0))
     {
       return TRUE;
@@ -985,9 +1007,4 @@ menu_proxy_module_load (UbuntuMenuProxyModule *module)
 G_MODULE_EXPORT void
 menu_proxy_module_unload (UbuntuMenuProxyModule *module)
 {
-  if (rebuild_ids)
-    {
-      g_hash_table_destroy (rebuild_ids);
-      rebuild_ids = NULL;
-    }
 }
